import socket
from .message import Message
from .message_event import MessageEvent
from .message_list import MessageList
from .connection import Connection
from .router import Router
from json_cpp import JsonList


class MessageClient:
    def __init__(self):
        self.failed_messages = None
        self.running = False
        self.registered = False
        self.router = Router()
        self.router.unrouted_message = self.__unrouted__
        self.ip = ""
        self.port = 0
        self.messages = MessageList()
        self.connection = None
        self._request_time_out = 500

    def __unrouted__(self, message: Message):
        self.messages.queue(message)

    def connect(self, ip: str, port: int):
        self.ip = ip
        self.port = port
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.ip, self.port))
        except socket.error as msg:
            return False
        self.connection = Connection(s, self.failed_messages)
        self.router.attend(self.connection)
        return True

    def send_request(self, message: Message, *args, **kwargs) -> Message:
        time_out = self._request_time_out
        if "time_out" in kwargs:
            time_out = kwargs["time_out"]
            kwargs.pop("time_out")
        if isinstance(message, Message):
            if args:
                if isinstance(args[0],int):
                    time_out = args[0]
                else:
                    raise RuntimeError("Wrong parameters")
        elif isinstance(message, str):
            if args:
                message = Message(message, args[0])
            else:
                if kwargs:
                    message = Message(message, **kwargs)
                else:
                    message = Message(message)
        else:
            raise RuntimeError("Wrong parameters")
        event = MessageEvent()
        self.router.add_message_event(request_id=message.id, event=event)
        self.send_message(message)
        response = event.wait(time_out)
        if response:
            return response
        raise TimeoutError("the request has timed_out")

    def set_request_time_out(self, time_out: int):
        self._request_time_out = time_out

    def send_async_request(self, message: Message, *args, **kwargs):
        call_back = None
        if "call_back" in kwargs:
            if callable(kwargs["call_back"]):
                call_back = kwargs["call_back"]
            else:
                raise RuntimeError("Wrong parameters")
            kwargs.pop("call_back")
        if isinstance(message, Message):
            if args:
                if callable(args[0]):
                    call_back = args[0]
                else:
                    raise RuntimeError("Wrong parameters")
        elif isinstance(message, str):
            if args:
                message = Message(message, args[0])
            else:
                if kwargs:
                    message = Message(message, **kwargs)
                else:
                    message = Message(message)
        else:
            raise RuntimeError("Wrong parameters")

        event = MessageEvent(call_back=call_back)
        self.router.add_message_event(request_id=message.id, event=event)
        self.send_message(message)

    def get_manifest(self):
        return self.send_request(Message("!manifest")).get_body(JsonList)

    def subscribe(self):
        return self.send_request(Message("!subscribe")).body == "success"

    def unsubscribe(self):
        return self.send_request(Message("!unsubscribe")).body == "success"

    def send_message(self, message: Message, *args, **kwargs):
        if isinstance(message, str):
            if args:
                message = Message(message, args[0])
            else:
                if kwargs:
                    message = Message(message, **kwargs)
                else:
                    message = Message(message)
        self.connection.send(message)

    def disconnect(self):
        self.running = False

    def __del__(self):
        self.disconnect()

    def __bool__(self):
        return self.connection is True



from setuptools import setup

setup(name='tcp-messages',description='create json services easily',author='german espinosa',author_email='germanespinosa@gmail.com',packages=['tcp_messages'],install_requires=['json-cpp>=1.0.78'],license='MIT',version='1.0.42',zip_safe=False)



from .util import check_type
import json
from json_cpp import JsonObject, JsonList
from uuid import uuid1



class Message(JsonObject):

    def __init__(self, header="", body="", **kwargs):
        JsonObject.__init__(self)
        self.header = header
        if not body and kwargs:
            body = JsonObject.load(json_dictionary_or_list=kwargs)
        self.body = str(body)
        self.id = str(uuid1())
        self._source = None

    def get_body(self, body_type: type = None):
        if body_type:
            if body_type is JsonObject or body_type is JsonList:
                return JsonObject.load(self.body)
            elif issubclass(body_type, JsonObject) or issubclass(body_type, JsonList):
                return body_type.parse(self.body)
            elif body_type is str:
                return self.body
            elif body_type is bool:
                return self.body == "success" or self.body == "true" or self.body == "1"
            else:
                return body_type(json.loads(self.body))
        else:
            return self.body

    def set_body(self, v):
        self.body = str(v)

    def reply(self, message):
        check_type(message, Message, "wrong type for message")
        if self._source:
            self._source.send(message)
            return True
        else:
            return False


class MessagePart(JsonObject):

    def __init__(self, header: str = "", body: str = "", message_id: str = "", seq: int = 0, parts: int = 1):
        self.header = header
        self.body = body
        self.id = message_id
        self.seq = seq
        self.parts = parts

    def to_message(self) -> Message:
        message = Message(header=self.header, body=self.body)
        message.id = self.id
        return message


class MessageParts(JsonList):

    def __init__(self, message: Message = None):
        JsonList.__init__(self, list_type=MessagePart)
        if message:
            parts = (len(message.body) // 1024) + 1
            for i in range(parts):
                part = MessagePart(header=message.header, body=message.body[i*1024:(i+1)*1024], message_id=message.id, seq=i, parts=parts)
                self.append(part)

    def join(self) -> Message:
        if len(self) == 0:
            return Message()
        message = self[0].to_message()
        for i in range(1, len(self)):
            message.body += self[i].body
        return message

    def is_ready(self) -> bool:
        if len(self) == 0:
            return False
        return self[0].parts == len(self)